/*********************************************************/
// MODULE:		Synchronizing FIFO
//
// FILE NAME:	afifo.v
// AUTHOR:		Bob Zeidman, Zeidman Consulting
// 
//
// DESCRIPTION:	This module defines a Synchronizing FIFO.
// The FIFO memory is implemented as a ring buffer. The read
// pointer points to the beginning of the buffer, while the
// write pointer points to the end of the buffer.
//
/*********************************************************/

// DEFINES

`define FIFO_DEPTH 5	// Depth of FIFO (number of bytes)
`define FIFO_HALF 3		// Half depth of FIFO
						// (this avoids rounding errors)
`define FIFO_BITS 4		// Number of bits required to
						// represent the FIFO size
`define FIFO_WIDTH 8	// Width of FIFO data

// TOP MODULE
module afifo(
// INPUTS
input				  	reset_n,	// Active low reset
input [`FIFO_WIDTH-1:0]	data_in, 	// Data input to FIFO
input					read_n,	 	// Read FIFO (active low)
input					write_n,  	// Write FIFO (active low)

// OUTPUTS
output reg [`FIFO_WIDTH-1:0]	data_out,	// FIFO data output
output						full,		// FIFO is full
output						empty,		// FIFO is empty
output						half,		// FIFO is half full
										// or more
output [`FIFO_BITS-1:0] count, // How many locations in the FIFO
                                // are occupied?
output reg underflow, overflow       
);

							// The FIFO memory.
reg	 [`FIFO_WIDTH-1:0]	fifo_mem[0:`FIFO_DEPTH];
							// FIFO read pointer points to
							// the location in the FIFO to
							// read from next
reg	 [`FIFO_BITS-1:0]	rd_pointer;
							// FIFO write pointer points to
							// the location in the FIFO to
							// write to next
reg	 [`FIFO_BITS-1:0]	wr_pointer;

reg flag;

// PARAMETERS

// ASSIGN STATEMENTS
assign count = (wr_pointer >= rd_pointer?
		wr_pointer - rd_pointer:
		`FIFO_DEPTH + wr_pointer - rd_pointer);
assign  full = (count == `FIFO_DEPTH) ? 1'b1 : 1'b0;
assign  empty = (count == 0) ? 1'b1 : 1'b0;
assign  half = (count >= `FIFO_HALF) ? 1'b1 : 1'b0;

// MAIN CODE

// Look at the falling edge of the read signal
always @(negedge read_n or negedge reset_n) begin
	if (~reset_n) begin
		// Reset the FIFO pointer
		rd_pointer = `FIFO_BITS'b0;
		underflow = 1'b0;
	end	// Check for FIFO underflow
	else if (empty) 
		underflow = 1'b1;
	else 
	begin
	underflow = 1'b0;
	// Output the data
	data_out <= fifo_mem[rd_pointer];

	// Increment the read pointer
	// Check if the read pointer has gone beyond the
	// depth of the FIFO. If so, set it back to the
	// beginning of the FIFO
	if (rd_pointer == `FIFO_DEPTH)
		rd_pointer <= `FIFO_BITS'b0;
	else
		rd_pointer <= rd_pointer + 1'b1;
	end
end

// Look at the falling edge of the write signal
always @(negedge write_n or negedge reset_n) begin
	if (~reset_n) begin
		// Reset the FIFO pointer
		wr_pointer = `FIFO_BITS'b0;
		overflow = 1'b0;
	end
	// Check for FIFO overflow
	else if (full) 
		overflow = 1'b1;
	else
	begin
	overflow = 1'b0;
	// Store the data
	fifo_mem[wr_pointer] <= data_in;
	// Increment the write pointer
	// Check if the write pointer has gone beyond the
	// depth of the FIFO. If so, set it back to the
	// beginning of the FIFO
	if (wr_pointer == `FIFO_DEPTH)
		wr_pointer <= `FIFO_BITS'b0;
	else
		wr_pointer <= wr_pointer + 1'b1;
	end
end
endmodule
